/*
	File:		CodeSignSample.cpp

	Contains:	Test application for Macintosh Code Signing

	Version:	Apple Data Security Services 1.0

	Copyright:	(c) 1999 by Apple Computer, Inc., all rights reserved.

	Change History (most recent first):

	To Do:
	
	Notes:		This application will demonstrate the use of the following
				File Signing API calls:
				
				*	Step											File Signing API call
				-------------------------------------------------------------------------------------
				*	A file will be examined for a signature.		SecMacHasSignature
				*	If it has one, it is removed					SecMacRemoveAllSignatures
				*	A signing key is obtained from the keychain		FindX509Certificates (Keychain API)
				*	The file is signed								SecMacSignFileSimple
				*	The file is verified							SecMacVerifyFileSimple
				*	Add a callback so dialogs are movable			SecMacAddCallback
				*	The results of the verification are shown		SecMacDisplaySigner
				*		Remove the callback							SecMacRemoveCallback
				
				If _USESIMPLESIGNING is set to 0, SecMacSignFile will be called instead of SecMacSignFileSimple.
				This code also uses SecMacLoadSigPound and demonstrates the use of a SecProgressCallbackUPP
				
				If _USESIMPLEVERIFYING is set to 0, SecMacVerifyFile will be called instead of SecMacVerifyFileSimple.
				This code also uses SecMacGetDefaultPolicyOIDs, SecMacFindSignatureToVerify and demonstrates the use of a SecProgressCallbackUPP

				Not shown in sample code: SecMacRemoveSignature
				
				This test assumes that a keychain exists, is unlocked, and contains a signing certificate
*/


#include <iostream>
#include <Resources.h>
#include <Strings.h>
#include <list.h>
#include <fonts.h>
#include <StandardFile.h>
#include <Sioux.h>
#include <OSUtils.h>
#include <Controls.h>
#include <Dialogs.h>

#include "FileSigning.h"

using namespace std;  //introduces namespace std

static int Test(void);
static bool SelectAFileToSign(FSSpec& theFile);
static void SetupSIOUX(void);
static void StandardMacInit(void);
static bool TestSecMacHasSignature(void);
static pascal OSStatus VerifyDialogEventCallbackProc(KCEvent keychainEvent,KCCallbackInfo *info);
static OSStatus FindSigningCertificate(KCItemRef& signingCertificate);

#define _USESIMPLESIGNING	1	// set to 1 to use SecMacSignFileSimple instead of SecMacSignFile
#define _USESIMPLEVERIFYING 1	// set to 1 to use SecMacVerifyFileSimple instead of SecMacVerifyFile

//------------------------------------------------------------------------
//			Code and definitions here are only used when calling
//			SecMacSignFile instead of SecMacSignFileSimple
//------------------------------------------------------------------------

#if !(_USESIMPLESIGNING || _USESIMPLEVERIFYING)

typedef struct
{
	bool called;
	bool signing;	// false means verifying
} ProgInfo;

static pascal OSStatus SIOUXSecProgressCallbackProc(SecProgressCallbackInfo* info,void* userContext);
static OSStatus TestSecMacSignFile(FSSpec *theFile,KCItemRef signingCertificate,SecOptions secOptions);
static OSStatus TestSecMacVerifyFile(FSSpec *theFile,SecOptions secOptions,SecSignerRef *signer);

#endif

//------------------------------------------------------------------------
//			Start of code
//------------------------------------------------------------------------

int main(void)
{	
	StandardMacInit();
	SetupSIOUX();

	Test();
	return 0;
}

static int Test(void)
{
	OSStatus err;
	FSSpec theFile;
	const char *pName=(char *)&theFile.name[1];	// for debug display
	const SecSignatureType signatureType=kSecSignatureTypeCMS;
	
	// Pick a file to sign
	if (!SelectAFileToSign(theFile))
		return 0;

	// A file will be examined for a signature
	bool hasSig=SecMacHasSignature(&theFile,signatureType);
	cout << "File \"" << pName << "\"";
	if (hasSig)
		cout << " has";
	else
		cout << " does not have";
	cout << " an Code Signing signature" << endl;
	
	// If it has a signature, it is removed
	if (hasSig)
	{
		err=SecMacRemoveAllSignatures(&theFile,signatureType);
		if (err!=noErr)
			cout << "Error removing signatures: " << err << endl;
	}
	
	// A signing key is obtained from the keychain (Keychain API)
	KCItemRef signingCert=NULL;
	err=FindSigningCertificate(signingCert);
	if (err!=noErr)
	{
		cout << "Error finding a signing certificate: " << err << endl;
		return err;		// none of the other tests will work, so exit
	}
	
	//------------------------------------------------------------------------
	//			Signing the file
	//------------------------------------------------------------------------
	
	// The file is signed
	SecOptions secOptions=kSecOptionProgress;
	
#if _USESIMPLESIGNING
	err=SecMacSignFileSimple(&theFile,signingCert,secOptions);
#else
	err=TestSecMacSignFile(&theFile,signingCert,secOptions);
#endif
	
	if (err)
		cout << "Error: " << err << endl;
	else
		cout << "File \"" << pName << "\" signed successfully" << endl;

	//------------------------------------------------------------------------
	//			Verifying the file
	//------------------------------------------------------------------------
	
	SecSignerRef signer=NULL;

#if _USESIMPLEVERIFYING
	// The file is verified
	err=SecMacVerifyFileSimple(&theFile,secOptions,kSecTrustPolicyCodeSigning,&signer);
#else
	err=TestSecMacVerifyFile(&theFile,secOptions,&signer);
#endif

	if (err)
		cout << "Error: " << err << " during verification" << endl;
	else
		cout << "File \"" << pName << "\" verified successfully" << endl;
	
	// Add a callback so dialogs are movable
	KCCallbackUPP inCallbackFunction=NewKCCallbackUPP(VerifyDialogEventCallbackProc);
	KCAddCallback(inCallbackFunction,kSystemEventKCEventMask,NULL);

	// The results of the verification are shown
	err=SecMacDisplaySigner(err,signer,true,&theFile);
	if (err!=noErr)
		cout << "Error displaying signer: " << err << endl;

	// Remove the callback
	KCRemoveCallback(inCallbackFunction);

	// Clean up
	if (signer)
		SecRelease(signer);
		
	return 0;
}

static OSStatus FindSigningCertificate(KCItemRef& signingCertificate)
{
	signingCertificate = NULL;
	
	CFMutableArrayRef certificateItems=NULL;
	OSStatus err=KCFindX509Certificates(NULL,NULL,NULL,kCertSearchSigningAllowed,&certificateItems);
	if (err==noErr)
		err=KCChooseCertificate(certificateItems,&signingCertificate,SecMacGetDefaultPolicyOIDs(kSecTrustPolicyCodeSigning),kPolicyKCStopOn);
	
//	if (certificateItems)
//		CFRelease(certificateItems);

	return err;
}
								  
static pascal OSStatus VerifyDialogEventCallbackProc(KCEvent /*keychainEvent*/,KCCallbackInfo *info)
{
	Boolean bx=SIOUXHandleOneEvent(&info->event);	// if using PowerPlant, could call UModalAlerts::ProcessModalEvent(info->theEvent)
	cout << "Idling..." << endl;
	return noErr;
}

//------------------------------------------------------------------------
//
//			Code for the "non-Simple" functions
//
//------------------------------------------------------------------------

#if !_USESIMPLESIGNING
static OSStatus TestSecMacSignFile(FSSpec *theFile,KCItemRef signingCertificate,SecOptions secOptions)
{
	// This function illustrates how to use helper functions to call SecMacSignFile
	OSStatus err;
	Handle sigH;
	SecProgressCallbackUPP progressProc;
	ProgInfo progInfo={0,};
	
	// Set up the handle to the 'sig#' resource
	err=SecMacLoadSigPound(theFile,&sigH);
	if (err!=noErr)
		return err;
		
	// Create the Universal ProcPtr for the progress routine
	progressProc=NewSecProgressCallbackProc(SIOUXSecProgressCallbackProc);
	
	// Set our fields in our local context structure
	progInfo.signing=true;
	
	// Sign the file
	err=SecMacSignFile(theFile,signingCertificate,sigH,secOptions,progressProc,&progInfo);
	
	// Dispose the Universal ProcPtr for the progress routine
	DisposeRoutineDescriptor(progressProc);
	
	return err;
}
#endif

#if !_USESIMPLEVERIFYING
static OSStatus TestSecMacVerifyFile(FSSpec *theFile,SecOptions secOptions,SecSignerRef *signer)
{
	// This function illustrates how to use helper functions to call SecMacVerifyFile
	OSStatus err;
	SecProgressCallbackUPP progressProc;
	ProgInfo progInfo={0,};
	SInt16 signatureToVerify;
	
	// Find the signature to verify
	err=SecMacFindSignatureToVerify(theFile,kSecSignatureTypeCMS,&signatureToVerify);
	if (err!=noErr)
	{
		cout << "Could not find a File Signing signature" << endl;
		return err;
	}
	cout << "A File Signing signature exists with a resource ID of " << signatureToVerify << endl;
		
	// Create the Universal ProcPtr for the progress routine
	progressProc=NewSecProgressCallbackProc(SIOUXSecProgressCallbackProc);
	
	// Set our fields in our local context structure
	progInfo.signing=false;
	
	// Verify the file
	err=SecMacVerifyFile(theFile,secOptions,progressProc,&progInfo,signatureToVerify,
			SecMacGetDefaultPolicyOIDs(kSecTrustPolicyCodeSigning),kSecStopOnPolicy,signer);
								 	
	// Dispose the Universal ProcPtr for the progress routine
	DisposeRoutineDescriptor(progressProc);
	
	return err;
}

#endif

#if !(_USESIMPLESIGNING || _USESIMPLEVERIFYING)
pascal OSStatus SIOUXSecProgressCallbackProc(SecProgressCallbackInfo* info,void* userContext)
{
/*
	UInt32 							version;
	Str255							fileName;
	UInt32							bytesProcessed;
	UInt32							totalBytes;
	UInt32							itemsRemainingToSign;
	UInt32							totalItemsToSign;
*/
	ProgInfo *progInfo=reinterpret_cast<ProgInfo *>(userContext);
	char *pname=reinterpret_cast<char *>(&info->fileName[1]);
	float percent=100*info->bytesProcessed/info->totalBytes;
	
	if (!progInfo->called)
	{
		// Only print this the first time we are called
		progInfo->called=true;
		cout << (progInfo->signing?"Signing":"Verifying") << " file \"" << pname << "\"" << endl;
		cout << "   Bytes to process: " << info->totalBytes << endl;
	}
	cout << "   Processed: " << info->bytesProcessed << " (" << percent << "%)" << endl;
	
	return noErr;
}
#endif

//------------------------------------------------------------------------
//
//			Standard utility functions
//
//------------------------------------------------------------------------

static void StandardMacInit(void)
{
	// Set up all the Mac environment stuff
	MaxApplZone();
	MoreMasters();
	InitGraf(&qd.thePort);
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(nil);
	InitCursor();
}

static bool SelectAFileToSign(FSSpec& theFile)
{
	SFReply myReply;
	SFTypeList myTypes;			// types of files to display
	short numTypes=-1;			// allow anything to be signed/verified
	Point where={40,40};
	Str255 prompt="\pSelect a signed file";

	SFGetFile (where,(const unsigned char *)&prompt,(RoutineDescriptor *)0L,numTypes,
		(const unsigned long *)myTypes,(RoutineDescriptor *)0L,(SFReply *)&myReply);

	if (!myReply.good)
		return 0;
	theFile.vRefNum=myReply.vRefNum;
	theFile.parID=0;
	::memcpy(theFile.name,myReply.fName,myReply.fName[0]+1);

	// set up fileName as C string for debug output
	UInt8 len=myReply.fName[0];
	myReply.fName[len+1]=0;
	return 1;
}

static void SetupSIOUX(void)
{
	SIOUXSettings.setupmenus       = true;
	SIOUXSettings.autocloseonquit = false;
	SIOUXSettings.asktosaveonclose = false;
}


